/*
 * Copyright (c) 2013, 2014, 2015, 2016
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: replicant.c 1261 2016-12-28 03:12:38Z leres $ (LBL)";
#endif

#include <sys/types.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "acld.h"

/* Locals */
static u_int32_t comment_recookie;
static int comment_readresponse;		/* true when reading comment */

/* Forwards */
static struct rereq *replicantfindrereq(struct state *,  u_int32_t);

void
replicantappendreq(struct rereq *rep, struct rereq **repp)
{
	struct rereq *rep2;

	if (*repp == NULL)
		*repp = rep;
	else {
		rep2 = *repp;
		while (rep2->next != NULL)
			rep2 = rep2->next;
		rep2->next = rep;
	}
}

void
replicantdone(struct state *sp, struct rereq *rep)
{
	struct client *cl;
	struct req *rp;

	/* Mark the replicant request as done */
	rep->state = RESTATE_DONE;

	/* We're done if this is a server request */
	if (rep->type == REQ_UNKNOWN || rep->type == REQ_REPLICANT)
		return;

	/* We're also done if the local request hasn't completed yet */
	if (rep->req != NULL && rep->req->state != RSTATE_DONE)
		return;

	/* Attempt to complete this client request */
	for (cl = sp->clients; cl != NULL; cl = cl->next) {
		for (rp = cl->req; rp != NULL; rp = rp->next) {
			if (rp->rereq == rep) {
				clientsend(cl, rp);
				return;
			}
		}
	}
	lg(LOG_ERR, "replicantdone: Can't find matching request");
	// exit(EX_SOFTWARE);
}

void
replicantfreerereq(struct state *sp, struct rereq *orep)
{
	struct rereq *rep, **repp;

	if (orep->aclname != NULL) {
		free(orep->aclname);
		orep->aclname = NULL;
	}
	if (orep->comment != NULL) {
		free(orep->comment);
		orep->comment = NULL;
	}
	if (orep->response != NULL) {
		free(orep->response);
		orep->response = NULL;
	}

	/* Unlink this replicant request */
	repp = &sp->new_rereq;
	rep = sp->new_rereq;
	while (rep != NULL) {
		if (rep == orep) {
			*repp = rep->next;
			break;
		}
		repp = &rep->next;
		rep = rep->next;
	}
	repp = &sp->sent_rereq;
	rep = sp->sent_rereq;
	while (rep != NULL) {
		if (rep == orep) {
			*repp = rep->next;
			break;
		}
		repp = &rep->next;
		rep = rep->next;
	}
	free(orep);
}

void
replicantinit(struct state *sp)
{
	sp->re_secs = 1;
	if (sp->cf->c_portreplicant > 0)
		timerset(&sp->t_re_connect, sp->re_secs);
	else
		timerreset(&sp->t_re_connect);
	timerreset(&sp->t_re_timeout);
}

static struct rereq *
replicantfindrereq(struct state *sp,  u_int32_t cookie)
{
	struct rereq *rep;

	for (rep = sp->sent_rereq; rep != NULL; rep = rep->next)
		if (rep->cookie == cookie)
			return (rep);
	return (NULL);
}

void
replicantinput(struct state *sp)
{
	int an;
	u_int32_t cookie;
	char *cp, *ep, **av;
	struct rereq *rep;
	struct iobuf *ip;

	ip = &sp->re_rbuf;

	while (iohaveline(ip)) {
		rep = NULL;
		if (comment_readresponse) {
			rep = replicantfindrereq(sp, comment_recookie);
			if (rep == NULL) {
				lg(LOG_ERR, "replicantinput: "
				    "Can't find rereq for recookie %d",
				    comment_recookie);
				    comment_readresponse = 0;
			}
		}

		/* Collect response lines */
		if (rep != NULL && rep->state == RESTATE_READRESPONSE) {
			if (!iocomment(&rep->response, ip)) {
				lg(LOG_DEBUG, "replicant response \"%s\"",
				    pretty(rep->response));

				/* Try to complete this request */
				replicantdone(sp, rep);

				/* Unlink */
				--sp->replicantdepth;
				if (sp->replicantdepth <= 0)
					timerreset(&sp->t_re_timeout);
				if (rep->req == NULL)
					replicantfreerereq(sp, rep);
				comment_readresponse = 0;
			}
			continue;
		}

		cp = iogetstr(ip);
		if (cp == NULL)
			break;

		/* Break into arguments */
		trimws(cp, strlen(cp));
		av = NULL;
		an = makeargv(cp, &av);
		if (an == 0) {
			lg(LOG_DEBUG, "replicantinput: blank input line");
			freeargv(av);
			continue;
		}
		if (an < 3 || an > 4) {
			lg(LOG_DEBUG, "replicantinput: garbage: \"%s\"",
			    pretty(cp));
			freeargv(av);
			continue;
		}

		/* Parse cookie */
		cookie = strtoul(av[1], &ep, 10);
		if (*ep != '\0') {
			lg(LOG_ERR, "replicantinput: non-numeric cookie: %s",
			    pretty(cp));
			freeargv(av);
			continue;
		}

		/* Find rereq that matches this cookie */
		rep = replicantfindrereq(sp, cookie);
		if (rep == NULL) {
			lg(LOG_ERR,
			    "replicantinput: Can't find rereq for cookie %d",
				cookie);
			freeargv(av);
			continue;
		}

		if (an == 4) {
			if (strcmp(av[an - 1], "-") != 0) {
				lg(LOG_ERR,
				    "replicantinput: garbage 2: %s",
				    pretty(cp));
				freeargv(av);
				continue;
			}
			rep->state = RESTATE_READRESPONSE;
			comment_recookie = cookie;
			++comment_readresponse;
			freeargv(av);
			continue;
		}

		if (rep->req != NULL && failed(rep->req->cmd, av[2]))
			rep->reflags |= REFLAG_FAILED;

		/* Try to complete this request */
		replicantdone(sp, rep);
		--sp->replicantdepth;
		if (sp->replicantdepth <= 0)
			timerreset(&sp->t_re_timeout);
		if (rep->req == NULL)
			replicantfreerereq(sp, rep);
		freeargv(av);
	}
}

void
replicantlogin(struct state *sp)
{
	int s;
	u_short port;
	socklen_t salen;
	struct cf *cf;
	struct rereq *rep, *rep2;
	struct addr *addr;
	struct sockaddr_storage sa;
	struct sockaddr *sap;
	char buf[64];

	cf = sp->cf;
	addr = &cf->c_replicantaddr;
	port = cf->c_portreplicant;
	sap = (struct sockaddr *)&sa;
	salen = addr2sa(addr, port, sap);
	(void)snprintf(buf, sizeof(buf), "%s.%hu", addr2str(addr), port);

	s = socket(addr->family, SOCK_STREAM, 0);
	if (s < 0) {
		lg(LOG_ERR, "socket: %s: %s", buf, strerror(errno));
		exit(EX_OSERR);
	}

	if (connect(s, sap, salen) < 0) {
		lg(LOG_ERR, "connect: %s: %s", buf, strerror(errno));
		(void)close(s);
		sp->re_secs = sp->re_secs * 2;
		if (sp->re_secs > 8)
			sp->re_secs = 8;
		timerset(&sp->t_re_connect, sp->re_secs);
		timerreset(&sp->t_re_timeout);
		return;
	}

	/* XX should this return status? */
	blocking(s, 0);
	sp->refd = s;
	sp->re_secs = 1;
	timerreset(&sp->t_re_connect);
	lg(LOG_INFO, "replicantlogin: connect");

	/* Toss stale I/O */
	iofree(&sp->re_rbuf);
	iofree(&sp->re_wbuf);

	/* Move sent replicant requests to front of new requests queue */
	if (sp->sent_rereq != NULL) {
		replicantappendreq(sp->new_rereq, &sp->sent_rereq);
		sp->new_rereq = sp->sent_rereq;
		sp->sent_rereq = NULL;
	}

	/* Remove stale acld and replicant requests */
	rep = sp->new_rereq;
	while (rep != NULL) {
		if (strcmp(rep->cmd, "acld") != 0 &&
		    strcmp(rep->cmd, "replicant") != 0)
			rep2 = NULL;
		else
			rep2 = rep;
		rep = rep->next;
		if (rep2 != NULL) {
			rep2->next = NULL;
			replicantfreerereq(sp, rep2);
		}
	}

	/* Fake acld request to match greeting */
	rep = new(1, sizeof(*rep), "replicantlogin: acld");
	rep->cmd = "acld";
	rep->type = REQ_UNKNOWN;
	rep->state = RESTATE_PENDING;
	rep->cookie = 0;		/* acld initial greeting cookie */
	rep->next = sp->sent_rereq;
	sp->sent_rereq = rep;
	++sp->replicantdepth;

	/* Identify ourselves before sending stale requests */
	rep = new(1, sizeof(*rep), "replicantlogin: replicant");
	rep->cmd = "replicant";
	rep->type = REQ_REPLICANT;
	rep->state = RESTATE_PENDING;
	rep->next = sp->new_rereq;
	sp->new_rereq = rep;
}

void
replicantprocess(struct state *sp)
{
	struct rereq *rep, **repp;

	/* Send as many new requests as we can */
	repp = &sp->new_rereq;
	rep = sp->new_rereq;
	while (rep != NULL && sp->replicantdepth <= REPLICANT_DEPTH) {
		replicantsend(sp, rep);
		repp = &rep->next;
		rep = rep->next;
	}

	/* Move the newly sent replicant requests to end of sent queue */
	*repp = NULL;
	replicantappendreq(sp->new_rereq, &sp->sent_rereq);
	sp->new_rereq = rep;
}

#ifdef notdef
void
replicantreport(struct state *sp, struct rereq *rep, const char *cmd)
{
	const char *p;
	char detail[132];

	switch (rep->type) {

	case REQ_DROP:
	case REQ_RESTORE:
		(void)snprintf(detail, sizeof(detail), "%s",
		    addr2str(&rep->acl.addr1));

	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_DROPTCPSYNPORT:
	case REQ_RESTORETCPSYNPORT:
		(void)snprintf(detail, sizeof(detail), "%s", rep->acl.port1);
		break;

	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
		(void)snprintf(detail, sizeof(detail), "%s %s",
		    addr2str(&rep->acl.addr1), rep->acl.port1);
		break;

	case REQ_NULLZERO:
	case REQ_NONULLZERO:
		(void)snprintf(detail, sizeof(detail), "%s",
		    addr2str(&rep->nullzero.dst));
		break;

	case REQ_ADDWHITELIST:
	case REQ_REMWHITELIST:
		(void)snprintf(detail, sizeof(detail), "%s",
		    addr2str(&rep->whitelist));
		break;

	case REQ_UNKNOWN:
	case REQ_REPLICANT:
		detail[0] = '\0';
		break;

	default:
		lg(LOG_ERR, "replicantreport: unknown type %s",
		    val2str(cmd2req, rep->type));
		abort();
	}

	if (failed(rep->cmd, cmd))
		p = " failed";
	else
		p = "";

	lg(LOG_INFO, "send to replicant@%s: %s %s%s",
	    addr2str(&sp->cf->c_replicantaddr), rep->cmd, detail, p);
}
#endif

void
replicantreqhook(struct state *sp, struct req *rp)
{
	struct rereq *rep;
	struct cf *cf;

	/* Check for interest request types */
	cf = sp->cf;
	switch (rp->type) {

	case REQ_DROP:
	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_DROPTCPSYNPORT:
	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_RESTORE:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_RESTORETCPSYNPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
		if (!cf->c_replicate_acl)
			return;
		break;

	case REQ_NULLZERO:
	case REQ_NONULLZERO:
		if (!cf->c_replicate_nullzero)
			return;
		break;

	case REQ_ADDWHITELIST:
	case REQ_REMWHITELIST:
		if (!cf->c_replicate_whitelist)
			return;
		break;

	default:
		return;
	}

	rep = new(1, sizeof(*rep), "replicantreqhook: rereq");
	rep->type = rp->type;
	rep->state = RESTATE_PENDING;
	rep->cmd = rp->cmd;
	if (rp->aclname != NULL)
		rep->aclname = strsave(rp->aclname);
	if (rp->comment != NULL)
		rep->comment = strsave(rp->comment);

	switch (rp->type) {

	case REQ_DROP:
	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_DROPTCPSYNPORT:
	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_RESTORE:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_RESTORETCPSYNPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
		rep->acl = rp->acl;
		break;

	case REQ_NULLZERO:
	case REQ_NONULLZERO:
		rep->nullzero = rp->nullzero;
		break;

	case REQ_ADDWHITELIST:
	case REQ_REMWHITELIST:
		rep->whitelist = rp->whitelist;
		break;

	default:
		abort();
	}

	/* Add cross-links */
	rp->rereq = rep;
	rep->req = rp;

	/* It's an immediate fail if the replicant isn't connected */
	if (sp->refd < 0) {
		rep->response = strsave("replicant not connected\r\n");
		rep->reflags |= REFLAG_FAILED;

		/* Try to complete this request */
		replicantdone(sp, rep);
		return;
	}

	/* Add new request to end of new replicant request queue */
	replicantappendreq(rep, &sp->new_rereq);
}

void
replicantsend(struct state *sp, struct rereq *rep)
{
	const char *post;

	if (rep->comment != NULL)
		post = " -";
	else
		post = "";

	switch (rep->type) {

	case REQ_DROP:
	case REQ_RESTORE:
		rep->cookie = ++sp->recookie;
		replicantwrite(sp, "%s %lu %s%s",
		    rep->cmd, rep->cookie, addr2str(&rep->acl.addr1), post);
		break;

	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_DROPTCPSYNPORT:
	case REQ_RESTORETCPSYNPORT:
		rep->cookie = ++sp->recookie;
		replicantwrite(sp, "%s %lu %s %s%s",
		    rep->cmd, rep->cookie, rep->acl.port1, rep->aclname, post);
		break;

	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
		rep->cookie = ++sp->recookie;
		replicantwrite(sp, "%s %lu %s %s%s", rep->cmd, rep->cookie,
		    addr2str(&rep->acl.addr1), rep->acl.port1, post);
		break;

	case REQ_NULLZERO:
	case REQ_NONULLZERO:
		rep->cookie = ++sp->recookie;
		replicantwrite(sp, "%s %lu %s%s",
		    rep->cmd, rep->cookie, addr2str(&rep->nullzero.dst), post);
		break;

	case REQ_ADDWHITELIST:
	case REQ_REMWHITELIST:
		rep->cookie = ++sp->recookie;
		replicantwrite(sp, "%s %lu %s%s",
		    rep->cmd, rep->cookie, addr2str(&rep->whitelist), post);
		break;

	case REQ_REPLICANT:
		rep->cookie = ++sp->recookie;
		replicantwrite(sp, "%s %lu%s", rep->cmd, rep->cookie, post);
		break;

	case REQ_UNKNOWN:
		/* Ignore greeting */
		return;

	default:
		lg(LOG_ERR, "replicantsend: unknown type %s",
		    val2str(cmd2req, rep->type));
		abort();
	}

	if (rep->comment != NULL) {
		replicantwrite(sp, "%s", rep->comment);
		replicantwrite(sp, ".");
	}
	rep->state = RESTATE_REPLICANT;
	++sp->replicantdepth;
	timerreset(&sp->t_re_timeout);
	timerset(&sp->t_re_timeout, sp->cf->c_replicant_session_secs);
}

void
replicantterminate(struct state *sp, const char *response)
{
	struct req *rp;
	struct rereq *rep;
	struct client *cl;

	if (sp->refd >= 0) {
		(void)close(sp->refd);
		sp->refd = -1;
	}
	sp->re_secs = 1;
	timerset(&sp->t_re_connect, sp->re_secs);
	iofree(&sp->re_rbuf);
	iofree(&sp->re_wbuf);

	/* Wrap up the replicant side of pending requests */
	for (rep = sp->new_rereq; rep != NULL; rep = rep->next) {
		strappend(&rep->response, response);
		strappend(&rep->response, "\r\n");
		rep->state = RESTATE_DONE;
		rep->reflags |= REFLAG_FAILED;
		rp = rep->req;
		if (rp != NULL) {
			cl = getreqclient(sp, rp);
			if (cl != NULL)
				clientsend(cl, rp);
		}
	}
	for (rep = sp->sent_rereq; rep != NULL; rep = rep->next) {
		strappend(&rep->response, response);
		strappend(&rep->response, "\r\n");
		rep->state = RESTATE_DONE;
		rep->reflags |= REFLAG_FAILED;
		rp = rep->req;
		if (rp != NULL) {
			cl = getreqclient(sp, rp);
			if (cl != NULL)
				clientsend(cl, rp);
		}
	}
	sp->replicantdepth = 0;
	timerreset(&sp->t_re_timeout);
}

/* Queue up output */
void
replicantwrite(struct state *sp, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	ioappendvfmt(&sp->re_wbuf, fmt, ap);
	va_end(ap);
}
